from backend.management_system.exceptions import Exceptions


def _get_lines(file_name):
    try:
        confLines = open(file_name, "rb").read().rstrip("\n").split("\n")
    except IOError as e:
        raise Exceptions.ManagementException("Failed to read vpn conf file - %s" % str(e))
    # Return data from file
    return confLines


def _find_marker(confLines, line_marker):
    for i, line in enumerate(confLines):
        if line.lower() == line_marker.lower():
            # Found marker
            markerLineIdx = i
            break
    else:
        # Didn't find the marker.
        raise Exceptions.ManagementException("Couldn't find the #Users marker line in the StrongSwan conf file")
    return markerLineIdx


def file_exist(file_name):
    try:
        _get_lines(file_name)
    except IOError as e:
        return False
    return True


def get_marker_index(file_name, line_marker):
    # Get lines
    confLines = _get_lines(file_name)
    # Find the '#Users' line, which marks the start of the users section.
    markerLineIdx = _find_marker(confLines, line_marker)
    # Return marker line index
    return markerLineIdx


def get_before_marker(file_name, line_marker, text=True):
    # Get lines
    confLines = _get_lines(file_name)
    # Find the '#Users' line, which marks the start of the users section.
    markerLineIdx = _find_marker(confLines, line_marker)
    # Return after marker
    data = []
    if len(confLines) > markerLineIdx + 1:
        data = confLines[0:markerLineIdx]
    if text:
        return "\n".join(data)
    return data


def get_after_marker(file_name, line_marker, text=True):
    # Get lines
    confLines = _get_lines(file_name)
    # Find the '#Users' line, which marks the start of the users section.
    markerLineIdx = _find_marker(confLines, line_marker)
    # Return after marker
    data = []
    if len(confLines) > markerLineIdx + 1:
        data = confLines[markerLineIdx + 1:]
    if text:
        return "\n".join(data)
    return data



"""This package implements real-time data collection capabilities on top of the python environment.

In order to use Rookout data collection service, the following components are needed:
- The Rook module package (this package).
- The Rookout Agent.
- The Rookout Web Service.

For further information, checkout rookout.com.

This package can be used in two ways:
- Importing using "from rook import auto_start" will start the Rook using the default configuration.
- Importing "from rook import Rook" allows to manually configure and start the Rook.
"""

# During development package structure is a bit different, so we extend the package dynamically
try:
    from . import lib
except ImportError:
    import os
    root_dir = os.path.dirname(os.path.dirname(os.path.dirname(os.path.dirname(os.path.abspath(__file__)))))
    __path__.append(root_dir)

# There appears to be issues if this code is loaded too late, so we load it early as possible
from .services.bdb_location_service import BdbLocationService as _BdbLocationService

from .interface import Rook



"""This is the external interface to the Rook package."""

# FUTURE - convert dcostring to restructured text


class Rook(object):
    """This class represents a control object for the module."""

    def __init__(self):
        """Initialize a new Rook controller."""
        import rook.singleton

        self._rook = rook.singleton.singleton_obj

    def start(self, host=None, port=None):
        """Start the rook module.

        Arguments:
        host - The address to the Rookout Agent. Only "localhost" is recommended.
        port - The port the Rookout Agent is listening on.
        """
        return self._rook.connect(host, port)

    def set_version_information(self, **kwargs):
        """Set the application's version information.

        If the agent connection has already been established, updated version information will be sent immediately.
        If the agent connection has not yet been established, the version information will be sent on connection.
        """
        self._rook.set_version_information(**kwargs)

    def collect_data(self, description, send_by_default=True, **kwargs):
        """Collect application data at this line.

        collect_data causes the following chain of events:
        - a data collection event is trigger based on the given description.
        - If there are Augs registered on this event:
            - Augs are provdided with the key word arguments are are executed.
        - If there are no Augs registered on this event:
            - If send_by_default is True a default Aug is created that sends a basic monitor message with the
                description and the key word arguments home.
            - If send_by_default is False nothing happens.
        """
        raise NotImplementedError()
        self._rook.collect_data(description, send_by_default, kwargs)



import fnmatch
import os
import shutil
from ..code_weld import BasicTextCodeWelder, WELD_POS
import logging


class FadingFiles(object):
    """
    Create files (and optionaly perform actions on them), and remove them on context exit
    """

    class FileOp(object):

        def __init__(self, to_remove=None):
            self._to_remove = to_remove or []

        def operate(self):
            pass

        def log_op(self):
            logging.warn("Performing %s" % self.__class__.__name__)

        def log_remove(self):
            logging.warn("Fading out files- %s for '%s'" % (self._to_remove, self.__class__.__name__))

        def remove(self):
            for path in self._to_remove:
                if os.path.isdir(path) and len(path) > 6:
                    shutil.rmtree(path)
                else:
                    os.remove(path)

    class CopyOp(FileOp):
        def __init__(self, src, dst):
            super(FadingFiles.CopyOp, self).__init__([dst])
            self.src = src
            self.dst = dst

        def log_op(self):
            logging.warn("FADING-COPY: '%s' -> '%s'" % (self.src, self.dst))

        def operate(self):
            shutil.copyfile(self.src, self.dst)

    class DumpOp(FileOp):
        def __init__(self, data, dst):
            super(FadingFiles.DumpOp, self).__init__([dst])
            self.dst = dst
            self.data = data

        def log_op(self):
            logging.warn("FADING-DUMP: Data['%s'... {%d}] -> '%s'" % (self.data[:10], len(self.data), self.dst))

        def operate(self):
            with open(self.dst, "wb") as dump:
                dump.write(self.data)

    class RevertOp(FileOp):
        def __init__(self, src):
            super(FadingFiles.RevertOp, self).__init__([])
            self.src = src
            self.data = None

        def operate(self):
            # backup
            with open(self.src, "rb") as org:
                self.data = org.read()

        def log_remove(self):
            logging.warn("Reverting (FadingFile): %s'" % self.src)

        def remove(self):
            # restore
            with open(self.src, "wb") as dump:
                dump.write(self.data)

    class InjectOp(RevertOp):
        def __init__(self, src, injection_rules):
            super(FadingFiles.InjectOp, self).__init__(src)
            self.rules = injection_rules

        def log_op(self):
            logging.warn("FADING-INJECTION: into '%s'" % self.src)

        def operate(self):
            super(FadingFiles.InjectOp, self).operate()
            with open(self.src, "wb") as target:
                weld = BasicTextCodeWelder(self.rules).weld(self.data)
                target.write(weld)

    def __init__(self, file_op_list, should_log=False):
        self._logging = should_log
        self.file_list = []
        for file_op in file_op_list:
            if isinstance(file_op, self.FileOp):
                self.file_list.append(file_op)
            else:
                self.file_list.append(self.FileOp([os.path.abspath(file_op)]))

    def __enter__(self):
        for op in self.file_list:
            if self._logging:
                op.log_op()
            op.operate()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        for op in self.file_list:
            if self._logging:
                op.log_remove()
            op.remove()


def collect_data_files(root_dir, directory_name, exclude=None):
    exclude = exclude or []
    directory = os.path.join(root_dir, directory_name)
    paths = []
    for (path, directories, file_names) in os.walk(directory):
        for filename in file_names:
            file_path = os.path.join(path[len(root_dir) + 1:], filename)
            if True not in [fnmatch.fnmatch(file_path, pattern) for pattern in exclude]:
                paths.append(file_path)
    return directory, paths



if "rook" in __name__:
    # This is caused by not changing sys.path in Rook combined with the way we package the code
    from rook.interfaces.config import LoggingConfiguration
else:
    from interfaces.config import LoggingConfiguration

from .logger_factory import LoggerFactory

def _build_logger():
    import sys
    if "unittest" in sys.argv[0]:
        log_level = 10
        log_to_stderr = True
    else:
        log_level = LoggingConfiguration.LOG_LEVEL
        log_to_stderr = LoggingConfiguration.LOG_TO_STDERR

    return LoggerFactory(
        LoggingConfiguration.LOGGER_NAME,
        log_level,
        LoggingConfiguration.FILE_NAME,
        log_to_stderr,
        LoggingConfiguration.LOG_TO_REMOTE,
        LoggingConfiguration.PROPAGATE_LOGS
    ).get_logger()

logger = _build_logger()
logger.info("Initialized logger")



import datetime
import re
import socket
import uuid
# With statement over Socket
from contextlib import contextmanager
from datetime import timedelta
from random import randrange, SystemRandom

__author__ = 'OrW'

ISO_TIMESTAMP_FORMAT = "%Y-%m-%dT%H:%M:%S.%f"


def datetime_from_iso_string(timestamp):
    format_string = ISO_TIMESTAMP_FORMAT
    if timestamp.endswith("Z"):
        # Not all ISO-8601 strings end with "Z".
        format_string += 'Z'
    return datetime.datetime.strptime(timestamp, format_string)


@contextmanager
def socket_context(*args, **kw):
    s = socket.socket(*args, **kw)
    yield s
    s.close()


class IpUtils(object):
    def __init__(self):
        self._own_ip = None
        self._ip_name_cache = None

    def get_own_ip(self):
        if self._own_ip is None:
            with socket_context(socket.AF_INET, socket.SOCK_DGRAM) as s:
                try:
                    s.connect(("google.com", 80))
                    self._own_ip = s.getsockname()[0]
                except Exception:
                    from config.config import logger
                    logger.warning("Failed to acquire external IP - binding server on 0.0.0.0")
                    self._own_ip = "0.0.0.0"
        return self._own_ip

    def get_ip_name(self, ip):
        if self._ip_name_cache is None:
            from expiringdict import ExpiringDict
            ExpiringDict(max_len=1000, max_age_seconds=200)
        try:
            res = self._ip_name_cache.get(ip, None)
            if res is None:
                # TODO: Remove patch.
                return None
                # res = socket.gethostbyaddr(ip)[0]
                # self._ip_name_cache[ip] = res
            return res or None
        except socket.herror:
            self._ip_name_cache[ip] = 0
            return None
        except socket.gaierror:
            self._ip_name_cache[ip] = 0
            return None

    @staticmethod
    def get_ip_for_dns_name(name):
        return socket.gethostbyname(name)


IP_UTILS = IpUtils()


class DomainUtils(object):
    def __init__(self, db):
        self._db = db

    def get_ip_names(self, ip):
        names = self._db.get_dns_names_for_ip(ip)
        if len(names) == 0:
            return [IP_UTILS.get_ip_name(ip)]
        else:
            return names

    def get_most_relevant_tld(self, ip):
        import tldextract
        names = self.get_ip_names(ip)
        tlds = {}
        # Get and count tld appearances
        for name in names:
            if isinstance(name, (str, unicode)) and len(name) > 0:
                tld = tldextract.extract(name).registered_domain
                tlds[tld] = tlds.get(tld, 0) + 1
        # find the tld that appeared the most
        if len(tlds) > 0:
            most_common_tld = max([(v, k) for k, v in tlds.iteritems()])[1]
        else:
            most_common_tld = None
        # return all the names, and their most relevant tld
        return names, most_common_tld


class RandomUtils(object):
    @staticmethod
    def gen_uid():
        return uuid.uuid4().hex

    @staticmethod
    def gen_token(size=256):
        if size % 2 != 0:
            raise ValueError("Size in bits must be an even number.")
        return uuid.UUID(int=SystemRandom().getrandbits(size/2)).hex + \
               uuid.UUID(int=SystemRandom().getrandbits(size/2)).hex


    @staticmethod
    def random_datetime(start=None, end=None):
        """
        This function will return a random datetime between two datetime
        objects.
        If no range is provided - a last 24-hours range will be used
        :param datetime,None start: start date for range, now if None
        :param datetime,None end: end date for range, next 24-hours if None
        """
        # build range
        now = datetime.datetime.now()
        start = start or now
        end = end or now + timedelta(hours=24)
        delta = end - start
        int_delta = (delta.days * 24 * 60 * 60) + delta.seconds
        # Random time
        random_second = randrange(int_delta)
        # Return random date
        return start + timedelta(seconds=random_second)


gen_uid = RandomUtils.gen_uid
gen_token = RandomUtils.gen_token


class StringUtils(object):
    @staticmethod
    def convert_camelcase_to_underscore(name, lower=True):
        s1 = re.sub('(.)([A-Z][a-z]+)', r'\1_\2', name)
        res = re.sub('([a-z0-9])([A-Z])', r'\1_\2', s1)
        if lower:
            return res.lower()
        else:
            return res.upper()



import os
import six
import inspect

from . import cdbg_native
from .module_explorer import GetCodeObjectAtLine

from rook.lib.logger import logger

from ..exceptions import RookBdbCodeNotFound, RookBdbSetBreakpointFailed

_flags = None
cdbg_native.InitializeModule(_flags)


class Bdb(object):

    def __init__(self):
        self.frame_behind = True

        self.fncache = {}

        self._cookies = {}

        self.user_line = None

    def set_trace(self):
        # Not needed
        pass

    def canonic(self, filename):
        if filename[0] == "<" and filename[-1] == ">":
            return filename
        canonic = self.fncache.get(filename)
        if not canonic:
            canonic = os.path.abspath(filename)
            canonic = os.path.normpath(canonic)
            self.fncache[filename] = canonic
        return canonic

    def ignore_current_thread(self):
        # Not needed
        pass

    def set_break(self, module, filename, lineno):
        status, code_object = GetCodeObjectAtLine(module, lineno)
        if not status:
            raise RookBdbCodeNotFound(code_object)

        # Install the breakpoint
        cookie = cdbg_native.SetConditionalBreakpoint(code_object, lineno, None, self._callback)
        if not cookie:
            raise RookBdbSetBreakpointFailed()

        self._cookies[filename, lineno] = cookie

    def clear_break(self, filename, lineno):
        pos = filename, lineno
        try:
            cookie = self._cookies[pos]
        except KeyError:
            return

        cdbg_native.ClearConditionalBreakpoint(cookie)
        del self._cookies[pos]

    def clear_all_breaks(self):
        for cookie in six.itervalues(self._cookies):
            cdbg_native.ClearConditionalBreakpoint(cookie)

        self._cookies = {}

    def close(self):
        pass

    def _callback(self, event, frame):

        if event != cdbg_native.BREAKPOINT_EVENT_HIT:
            logger.error("Unexpected event- %d", event)
            return

        if frame and self.user_line:
            self.user_line(frame)



import shlex
import subprocess

import psutil

from lib import logger


class ExecutableProcess(object):
    """
    A Wrapper for process information (via psutil) and execution (via subprocess.Popen)
    """

    class ExecutableProcessError(Exception):
        pass

    def __init__(self, args, popen_kwargs=None):
        if args[0] == "start":
            self._check_args = args[1:]
        else:
            self._check_args = args
        # Popen values
        self._args = args
        self._kwargs = popen_kwargs or {}
        # The Popen object opened at execution
        self._popen_process = None
        self._process = None
        self.match_running()

    @classmethod
    def from_pid(cls, pid):
        """
        Create an ExecutableProcess from a given process id
        :param pid: the sought process id as int
        :return: ExecutableProcess or None on failure
        """
        if psutil.pid_exists(pid):
            proc = psutil.Process(pid)
            return cls(proc.cmdline(), popen_kwargs={'executable': proc.exe(), 'cwd': proc.cwd()})
        return None

    @classmethod
    def from_cmdline(cls, cmdline):
        """
        Create an ExecutableProcess from a given cmdline
        :param cmdline: the command line of the sought process as a string
        :return: ExecutableProcess or None on failure
        """
        encountered_access_denied = False
        cmdlist = cmdline.split(" ")
        for proc in psutil.process_iter():
            try:
                if proc.cmdline() == cmdlist:
                    return cls(proc.cmdline(), popen_kwargs={'executable': proc.exe(), 'cwd': proc.cwd()})
            except psutil.AccessDenied:
                encountered_access_denied = True
                # ignore processes we can't access
                continue
        if encountered_access_denied:
            # This warning doesn't mean that there's a problem; only that the WatchDog is running not as root/admin
            logger.logger.warning("%s.from_cmdline didn't find a process with cmdline:'%s'."
                                  "We might have missed the process due to AccessDenied errors" % (
                                      cls.__name__, cmdline))
        return None


    @classmethod
    def from_exec(cls, executable):
        """
        Create an ExecutableProcess from a given cmdline
        :param executable: the path to the program being run
        :return: ExecutableProcess or None on failure
        """
        encountered_access_denied = False
        for proc in psutil.process_iter():
            try:
                if proc.exe() == executable:
                    return cls(proc.cmdline(), popen_kwargs={'executable': proc.exe(), 'cwd': proc.cwd()})
            except psutil.AccessDenied:
                encountered_access_denied = True
                # ignore processes we can't access
                continue
        if encountered_access_denied:
            # This warning doesn't mean that there's a problem; only that the WatchDog is running not as root/admin
            logger.logger.warning("%s.from_cmdline didn't find a process with executable:'%s'."
                                  "We might have missed the process due to AccessDenied errors" % (
                                      cls.__name__, executable))
        return None


    @property
    def popen(self):
        return self._popen_process

    @property
    def info(self):
        return self._process

    def match_running(self):
        """
        Find a process object (representing a live process) matching our execution details
        :return a psutil.Process matching the execution args / kwargs provided, or None if not found.
        """
        # If we have a process object (Popen) we opened, check if it's still running
        if self._popen_process is not None and isinstance(self._process, psutil.Process):
            if self._process.is_running():
                return self._process
        # If we didn't open a process, or couldn't find our process running
        # Check if we can find a new matching process by execution details
        for process in psutil.process_iter():
            try:
                # if this process matches our execution details
                if process.cmdline() == self._check_args:
                    executable = self._kwargs["executable"]
                    if executable is not None and executable == process.exe():
                        self._process = process
                        return process
                    elif executable is None:
                        self._process = process
                        return process
            except Exception:
                continue
        # If we couldn't match a process, return None
        return None

    def suspend(self):
        assert isinstance(self._process, psutil.Process)
        self._process.suspend()

    def resume(self):
        assert isinstance(self._process, psutil.Process)
        self._process.resume()

    def kill(self):
        assert isinstance(self._process, psutil.Process)
        self._process.kill()

    def assure_running(self):
        """
        Check if a process like ours is running, if not start it
        """
        running = self.match_running()
        if running is None:
            self.execute()
        if not self.match_running():
            raise RuntimeError("Failed to start process %s" % self.info)

    def execute(self):
        """
        Start the process we are configured to, and update process info internally.
        """
        self._popen_process = subprocess.Popen(self._args, **self._kwargs)
        self._process = psutil.Process(self._popen_process.pid)

    @classmethod
    def start_process(cls, cmdline, executable=None, cwd=None,
                      shell=False,
                      execute_once=True, wait=False,
                      expect_code=0, expect_stdout=None, expect_stderr=""):
        """
        Start a sub-process
        :param [str,list] cmdline: command line to start the process
        :param str executable: optional executable path to be passed Popen
        :param str cwd: optional current working directory path to be passed to Popen
        :param bool shell: if True the cmdline will be passed to the shell
        :param bool execute_once: if True the process will not be executed if an equal one is already up
        :param bool wait: Should the call wait for the process to exit (uses Popen.communicate)
                          Enables usage of the expect_* arguments
        :param str expect_stderr: Used with wait=True, checks the resulting stderr
                                    and raises ExecutableProcessError if it differs from the passed value
        :param str expect_stdout: Used with wait=True, checks the resulting stdout
                                    and raises ExecutableProcessError if it differs from the passed value
        :param int expect_code: Used with wait=True, checks the resulting return-code
                                    and raises ExecutableProcessError if it differs from the passed value
        :returns: the ExecutableProcess of the started process
        """
        if isinstance(cmdline, list):
            args = cmdline
        else:
            args = shlex.split(cmdline)
        proc = cls(args=args, popen_kwargs={'executable': executable, 'cwd': cwd,
                                            'shell': shell,
                                            # Channel pipes so the process won't hang
                                            "stdin": subprocess.PIPE,
                                            "stdout": subprocess.PIPE,
                                            "stderr": subprocess.PIPE})
        if execute_once:
            proc.assure_running()
        else:
            proc.execute()
        if wait:
            stdout, stderr = proc.popen.communicate()
            if expect_code and expect_code != proc.popen.returncode:
                raise cls.ExecutableProcessError(
                    "Expected return-code %d for command '%s', got %d"
                    % (expect_code, cmdline, proc.popen.returncode))
            if expect_stderr and expect_stderr != stderr:
                raise cls.ExecutableProcessError(
                    "Expected stderr '%s' for command '%s', got '%s'"
                    % (expect_stderr, cmdline, stderr))
            if expect_stdout and expect_stdout != stdout:
                raise cls.ExecutableProcessError(
                    "Expected stdout '%s' for command '%s', got '%s'"
                    % (expect_stdout, cmdline, stdout))
        return proc

    @classmethod
    def run_shell(cls, cmdline, cwd=None, wait=False,
                  expect_code=0, expect_stdout=None, expect_stderr=""):
        """
        Convenience wrapper for start_process to run commands through the shell
        """
        proc = cls.start_process(cmdline, cwd=cwd, wait=wait, shell=True, execute_once=False,
                                 expect_code=expect_code, expect_stdout=expect_stdout,
                                 expect_stderr=expect_stderr)
        return proc



import socket
from subprocess import Popen, PIPE


class ShellUtils(object):
    """
    Utilities to run classic shell commands
    currently bash (unix) oriented
    """

    class ParsingException(Exception):
        pass

    @staticmethod
    def exec_cmd(cmd):
        """
        :param cmd: comamnd-line
        :return: (stdout, stderr) tuple
        """
        return Popen(cmd.split(" "), stdout=PIPE, stderr=PIPE).communicate()

    @classmethod
    def get_pid_of_process_name(cls, name):
        try:
            return int(cls.exec_cmd("pidof %s" % name)[0])
        except ValueError:
            return None

    @classmethod
    def get_process_connections(cls, process_name, only_listening=False):
        """
        Netstat parser
        """
        import psutil
        res = []
        pid = cls.get_pid_of_process_name(process_name)
        for connection in psutil.net_connections():
            if connection.pid == pid:
                if not only_listening or\
                                connection.status == "LISTEN" or\
                                connection.type != socket.SOCK_STREAM:
                    res.append(connection)
        return res






import json
import threading
import time
import six
import importlib

from grpc import insecure_channel, RpcError, StatusCode, secure_channel,\
    ssl_channel_credentials

from rook.lib.logger import logger

from ..services.bdb_location_service import BdbLocationService

from rook.interfaces.protobuf import rook_pb2_grpc
from rook.interfaces.protobuf import rook_pb2

from rook.interfaces.config import AgentComConfiguration

from rook.exceptions import RookCommunicationException


class AgentCom(object):
    """This class handles the communication itself."""

    def __init__(self, aug_manager, trigger_services, output, agent_host, agent_port, token):
        self._aug_manager = aug_manager
        self._bdb = trigger_services.get_service(BdbLocationService.NAME)
        self._output = output

        self.agent_host = agent_host
        self.agent_port = agent_port
        self.token = token

        self._channel = None
        self._client = None

        self._retries = 0

    def connect_to_agent(self):
        logger.debug("Attempting connection to agent")

        self.disable_gevent_for_grpc()
        self._build_client()

        try:
            self._start_new_connnection()
        except RookCommunicationException:
            logger.warn("Failed to connect to agent. Will keep trying")
            self._start_retry_thread()
            raise

    def send_rule_status_update(self, status_update):
        self._client.send_rule_status_update(status_update, metadata=self._build_metadata())

    def send_rook_messages(self, messages):
        self._client.send_messages(messages, metadata=self._build_metadata())

    def _build_client(self):
        if self._client:
            return

        logger.debug("Connecting to agent-\t%s:%d", self.agent_host, self.agent_port)
        options = [('grpc.max_receive_message_length', AgentComConfiguration.MAX_MESSAGE_LENGTH)]

        connection_string = self.agent_host + ':' + str(self.agent_port)

        if self.token:
            credentials = ssl_channel_credentials()
            self._channel = secure_channel(connection_string, credentials)
        else:
            self._channel = insecure_channel(connection_string, options=options)

        logger.debug("Creating client")
        self._client = rook_pb2_grpc.AgentManagementServiceStub(self._channel)

    def _build_metadata(self):
        metadata = []

        if self.token:
            metadata.append(("token", self.token))

        return metadata

    def _start_new_connnection(self):
        logger.debug("Starting a new connection")

        logger.debug("Getting rook info")
        info = self._output.get_rook_info()

        logger.debug("Pinging")
        ping_message = rook_pb2.PingMessage()
        ping_message.value = 1
        try:
            self._client.ping(ping_message, timeout=AgentComConfiguration.GRPC_TIMEOUT, metadata=self._build_metadata())
        except Exception as e:
            message = "Failed to ping agent"
            logger.info(message, exc_info=1)
            six.raise_from(RookCommunicationException(message), e)

        logger.debug("Creating request")
        command_request = self._client.new_rook_notification(info, metadata=self._build_metadata())

        # Get initial augs
        success = self._command_loop(command_request)
        if not success:
            raise RookCommunicationException("Failed to connect to agent")

        # Setup update/retry channel in the background
        logger.info("Creating commands thread")
        thread = threading.Thread(target=self._command_thread, args=(command_request, ),
                                  name=AgentComConfiguration.COMMAND_THREAD_NAME)
        thread.daemon = True
        thread.start()

    def _command_loop(self, commands_request):
        logger.debug("Entering command loop")

        try:
            for command in commands_request:
                self._retries = 0

                if command.command_type == rook_pb2.RookCommand.COMMAND_ADD_AUG:
                    logger.info("Got an add aug command")
                    try:
                        self._aug_manager.add_aug(json.loads(command.aug_json))
                    except:
                        logger.exception("Error processing new aug")

                elif command.command_type == rook_pb2.RookCommand.COMMAND_REMOVE_AUG:
                    logger.info("Got a remove aug command")
                    try:
                        self._aug_manager.remove_aug(command.aug_id)
                    except:
                        logger.exception("Error removing aug")

                elif command.command_type == rook_pb2.RookCommand.COMAND_INIT_FINISHED:
                    logger.info("Finished initialization")
                    self._output.start_sending_messages()
                    return "Init Finished"

                elif command.command_type == rook_pb2.RookCommand.COMMAND_CLEAR_AUGS:
                    logger.info("Got clear augs command")
                    try:
                        self._aug_manager.clear_augs()
                    except:
                        logger.exception("Error in clearing augs")

                elif command.command_type == rook_pb2.RookCommand.COMMAND_SET_ROOK_ID:
                    logger.info("Got set rook_id command- %s", command.rook_id)
                    self._output.set_rook_id(command.rook_id)

                else:
                    logger.error("Unknown command %s", command)
        except RpcError as exc:
            # This exception can occur during process shutdown
            if StatusCode.CANCELLED == exc.code():
                self._output.stop_sending_messages()
                return "Process Exiting"

            if self._retries:
                logger.info("Error in command loop", exc_info=1)
            else:
                logger.exception("Error in command loop")
        except:
            if self._retries:
                logger.info("Error in command loop", exc_info=1)
            else:
                logger.exception("Error in command loop")

        self._output.stop_sending_messages()
        logger.warn("Exiting with failure")
        return None

    def _command_thread(self, command_request):
        logger.info("Command thread started")

        # Disable bdb monitoring of this thread
        self._bdb.ignore_current_thread()

        if command_request:
            ret = self._command_loop(command_request)
            # True here means the process is exiting
            if ret:
                return

        self._start_retry_thread()

    def _start_retry_thread(self):
        logger.info("Retry thread started")
        thread = threading.Thread(target=self._retry_thread,
                                  name=AgentComConfiguration.COMMAND_THREAD_NAME)
        thread.daemon = True
        thread.start()

    def _retry_thread(self):
        logger.info("Starting retry thread")

        # Disable bdb monitoring of this thread
        self._bdb.ignore_current_thread()

        self._retries = 0

        while True:
            try:
                self._start_new_connnection()
                break
            except:
                # Interpreter shutdown
                if not logger:
                    return

                if self._retries > 0:
                    backoff = AgentComConfiguration.BACK_OFF * (2 ** (self._retries - 1))
                    sleep_time = min((backoff, AgentComConfiguration.MAX_SLEEP))
                    logger.info("Sleeping for- %f", sleep_time)
                    time.sleep(sleep_time)
                self._retries += 1

    def disable_gevent_for_grpc(self):
        try:
            import gevent.monkey
        except ImportError:
            return

        if not gevent.monkey.is_module_patched('threading'):
            return

        # Rebuild a clean module based on a bundled copy of threading
        from . import threading as clean_threading
        clean_threading._start_new_thread = gevent.monkey.get_original('thread', 'start_new_thread')
        clean_threading._allocate_lock = gevent.monkey.get_original('thread', 'allocate_lock')
        clean_threading._get_ident = gevent.monkey.get_original('thread', 'get_ident')
        clean_threading.stack_size = gevent.monkey.get_original('thread', 'stack_size')
        clean_threading.local = gevent.monkey.get_original('thread', '_local')

        MODULE_LIST = [
            'grpc.beta.implementations',
            'grpc.beta.utilities',
            'grpc.beta._server_adaptations',
            'grpc.framework.foundation.stream_util',
            'grpc.framework.interfaces.base.base',
            'grpc._channel',
            'grpc._common',
            'grpc._plugin_wrapping',
            'grpc._server',
            'grpc._utilities'
        ]

        for module_name in MODULE_LIST:
            module = importlib.import_module(module_name)
            module.threading = clean_threading

        # CleanupThread has already inherited, so we override it
        # Copied from gRPC in order to build a clean version of the class
        class CleanupThread(clean_threading.Thread):
            """A threading.Thread subclass supporting custom behavior on join().

            On Python Interpreter exit, Python will attempt to join outstanding threads
            prior to garbage collection.  We may need to do additional cleanup, and
            we accomplish this by overriding the join() method.
            """

            def __init__(self, behavior, *args, **kwargs):
                """Constructor.

                Args:
                    behavior (function): Function called on join() with a single
                        argument, timeout, indicating the maximum duration of
                        `behavior`, or None indicating `behavior` has no deadline.
                        `behavior` must be idempotent.
                    args: Positional arguments passed to threading.Thread constructor.
                    kwargs: Keyword arguments passed to threading.Thread constructor.
                """
                super(CleanupThread, self).__init__(*args, **kwargs)
                self._behavior = behavior

            def join(self, timeout=None):
                start_time = time.time()
                self._behavior(timeout)
                end_time = time.time()
                if timeout is not None:
                    timeout -= end_time - start_time
                    timeout = max(timeout, 0)
                super(CleanupThread, self).join(timeout)

        # Restore threading for _common
        import grpc._common
        grpc._common.CleanupThread = CleanupThread



import sys

from ..lib_interfaces import rook_pb2, ToolException, ErrorConfiguration
from .namespace_serializer import NamespaceSerializer
from .namespaces.python_object_namespace import PythonObjectNamespace
from .namespaces.stack_namespace import StackNamespace


class Error(object):

    def __init__(self, exc=None, type=None, message=None, parameters=None, traceback=None):
        self.exc = exc

        if isinstance(exc, ToolException):
            self.type = exc.get_type()
            self.message = exc.get_message()
            self.parameters = exc.get_parameters()
        else:
            self.type = type or "Unknown"
            self.message = message
            self.parameters = parameters

        # Extract traceback if was not supplied
        self.traceback = traceback
        if self.traceback is None and sys.exc_info()[2]:
            self.traceback = sys.exc_info()[2].tb_frame

    def dump(self, error):
        try:
            if self.message:
                error.message = NamespaceSerializer.normalize_string(self.message)
            else:
                error.message = ''
        except Exception:
            pass

        try:
            error.type = NamespaceSerializer.normalize_string(self.type)
        except Exception:
            pass

        try:
            serializer = NamespaceSerializer()
            serializer.dump(PythonObjectNamespace(self.parameters), error.parameters, False)
        except Exception:
            pass

        try:
            serializer = NamespaceSerializer()
            serializer.dump(PythonObjectNamespace(self.exc), error.exc, False)
        except Exception:
            pass

        try:
            if self.traceback:
                serializer = NamespaceSerializer()
                serializer.dump(StackNamespace(self.traceback).traceback(ErrorConfiguration.STACK_DEPTH), error.traceback, False)
        except Exception:
            pass

    def dumps(self):
        error = rook_pb2.Error()
        self.dump(error)
        return error



#!/usr/bin/env python

import os
import shutil
import platform
import sys
import warnings
import json
from glob import glob

try:
    # Use setuptools if available, for install_requires (among other things).
    import setuptools
    from setuptools import setup, Extension
except ImportError:
    setuptools = None
    from distutils.core import setup, Extension

# Classic setup.py

kwargs = {}

# Version configuration
with open('rook/rookout-config.json', 'r') as f:
    config = json.load(f)
version = config['VersionConfiguration']['VERSION']


# Readme
with open('README.rst') as f:
    kwargs['long_description'] = f.read()

# Extension modules
try:
    USE_CYTHON = os.environ['USE_CYTHON']
except:
    USE_CYTHON = 'sdist' in sys.argv or 'test' in sys.argv

if 'CPython' == platform.python_implementation():
    ext_modules = []

    if 'Linux' in platform.system() and platform.python_version().startswith('2.7'):
        ext_modules.append(Extension(
            'rook.services.cdbg_native',
            sources=glob('rook/services/native/*.cc'),
            extra_compile_args=[
                '-std=c++0x',
                '-Werror',
                '-g0',
                '-O3',
            ],
            libraries=['rt']))

    ext = '.pyx' if USE_CYTHON else '.c'
    cython_extensions = [Extension("rook.services.pyx_bdb", ["rook/services/pyx_bdb"+ext])]
    if USE_CYTHON:
        from Cython.Build import cythonize
        cython_extensions = cythonize(cython_extensions)
    ext_modules += cython_extensions

    kwargs['ext_modules'] = ext_modules

if setuptools is not None:
    # If setuptools is not available, you're on your own for dependencies.
    install_requires = [
        "six>=1.9",
        "grpcio==1.4.0",
    ]

    kwargs['install_requires'] = install_requires

setup(
    name="rook",
    version=version,
    packages=setuptools.find_packages(where='.', exclude=['contrib', 'docs', '*test*']),
    include_package_data=True,
    author="Rookout",
    author_email="liran@rookout.com",
    url="http://rookout.com/",
    description="Rook is a Python package for on the fly debugging and data extraction for application in production",
    classifiers=[
        'Development Status :: 3 - Alpha',

        'Intended Audience :: Developers',
        'Intended Audience :: System Administrators',

        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: Implementation :: CPython',

        ],
    zip_safe=False,
    test_suite='tests',
    **kwargs
)




import inspect
import six

from .namespace import Namespace
from .container_namespace import ContainerNamespace
from .python_object_namespace import PythonObjectNamespace

from ...lib_interfaces import RookAttributeNotFound


class FrameNamespace(Namespace):

    def __init__(self, frame):
        super(FrameNamespace, self).__init__(self.METHODS)

        self.frame = frame
        self.frame_info = inspect.getframeinfo(frame, 0)

    def read_attribute(self, name):
        if name in self.frame.f_locals:
            return PythonObjectNamespace(self.frame.f_locals[name])
        elif name in self.frame.f_globals:
            return PythonObjectNamespace(self.frame.f_globals[name])
        else:
            raise RookAttributeNotFound(name)

    def filename(self, args=None):
        return PythonObjectNamespace(self.frame_info[0])

    def line(self, args=None):
        return PythonObjectNamespace(self.frame_info[1])

    def function(self, args=None):
        return PythonObjectNamespace(self.frame_info[2])

    def module(self, args=None):
        module = inspect.getmodule(self.frame)
        if module:
            return PythonObjectNamespace(module.__name__)
        else:
            return PythonObjectNamespace(None)

    def locals(self, args=None):
        result = {}
        for name, value in six.iteritems(self.frame.f_locals):
            ns = PythonObjectNamespace(value)
            if name == 'self':
                ns.dump_config = PythonObjectNamespace.ObjectDumpConfig.conservative_limits(value)
            result[name] = ns

        return ContainerNamespace(result)

    def globals(self, args):
        return PythonObjectNamespace(self.frame.f_globals)

    def dump(self, args):
        return ContainerNamespace({
            'locals': self.locals(),
            'module': self.module(),
            'filename': self.filename(),
            'line': self.line(),
            'function': self.function(),
        })

    METHODS = (filename, line, function, module, locals, globals, dump)



"""This module initializes the Rook using default settings when imported."""
import six

from rook.interface import Rook
from rook.exceptions import RookCommunicationException

try:
    obj = Rook()
    obj.start()
except RookCommunicationException:
    six.print_("Rook failed to connect to the agent - will continue attempting in the background.")
    import traceback
    traceback.print_exc()
except BaseException:
    six.print_("Rook failed automatic initialization")
    import traceback
    traceback.print_exc()
